#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import argparse

import flask
from blockchain.block.create_block import CreateBlock
from blockchain.block.get_block import GetBlockFromOtherNode
from flask import jsonify
from flask import request
from lib.export import export_the_transactions
from lib.log import get_logger
from lib.settings_system import debug_mode
from lib.settings_system import test_mode
from lib.settings_system import the_settings
from lib.status import Status
from node.node import Node
from node.node_connection import Node_Connection
from node.unl import Unl
from transactions.get_my_transaction import GetMyTransaction
from transactions.send import send
from waitress import serve
from wallet.create_a_wallet import create_a_wallet
from wallet.delete_current_wallet import delete_current_wallet
from wallet.print_balance import print_balance
from wallet.print_wallets import print_wallets
from wallet.wallet_selector import wallet_selector

logger = get_logger("API")

app = flask.Flask(__name__)


@app.route("/wallet/print", methods=["GET"])
def print_wallets_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    return jsonify(print_wallets())


@app.route("/wallet/change/<number>", methods=["GET"])
def wallet_change_page(number):
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    wallet_selector(number)
    return jsonify(print_wallets())


@app.route("/wallet/create/<password>", methods=["GET"])
def create_wallet_page(password):
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    create_a_wallet(password)
    return jsonify(print_wallets())


@app.route("/wallet/delete", methods=["GET"])
def delete_wallets_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    delete_current_wallet()
    return jsonify(print_wallets())


@app.route("/send/coin/<address>/<amount>/<password>", methods=["GET"])
def send_coin_page(address, amount, password):
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    send(password, address, amount)
    return jsonify("OK")


@app.route("/send/coin-data/<address>/<amount>/<data>/<password>",
           methods=["GET"])
def send_coin_data_page(address, amount, data, password):
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    send(password, address, amount, data)
    return jsonify("OK")


@app.route("/wallet/balance", methods=["GET"])
def balance_wallets_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    return jsonify(print_balance())


@app.route("/node/start/<ip>/<port>", methods=["GET"])
def node_start_page(ip, port):
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    Node(str(ip), int(port))
    return jsonify("OK")


@app.route("/node/stop", methods=["GET"])
def node_stop_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    Node.main_node.stop()
    return jsonify("OK")


@app.route("/node/connect/<ip>/<port>", methods=["GET"])
def node_connect_page(ip, port):
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    Node_Connection.connect(str(ip), int(port))
    return jsonify("OK")


@app.route("/node/connectmixdb", methods=["GET"])
def node_connectmixdb_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    Node_Connection.connectmixdb()
    return jsonify("OK")


# /node/newunl/?MFYw......
@app.route("/node/newunl/", methods=["GET"])
def node_newunl_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    Unl.save_new_unl_node(request.query_string.decode("utf-8"))
    return jsonify("OK")


@app.route("/node/id", methods=["GET"])
def node_id_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    return jsonify(Node.id)


@app.route("/settings/test/on", methods=["GET"])
def settings_test_on_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    test_mode(True)
    return jsonify("OK")


@app.route("/settings/test/off", methods=["GET"])
def settings_test_off_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    test_mode(False)
    return jsonify("OK")


@app.route("/settings/debug/on", methods=["GET"])
def settings_debug_on_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    app.config["DEBUG"] = True
    debug_mode(True)
    return jsonify("OK")


@app.route("/settings/debug/off", methods=["GET"])
def settings_debug_off_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    app.config["DEBUG"] = False
    debug_mode(False)
    return jsonify("OK")


@app.route("/block/get", methods=["GET"])
def block_get_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    if the_settings()["test_mode"]:
        CreateBlock()
    else:
        GetBlockFromOtherNode()
    return jsonify("OK")


@app.route("/export/transactions/csv", methods=["GET"])
def export_transaction_csv_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    if export_the_transactions():
        return jsonify("OK")
    else:
        return jsonify("You have not a transaction")


@app.route("/export/transactions/json", methods=["GET"])
def export_transaction_json_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    return jsonify([i.__dict__ for i in GetMyTransaction()])


@app.route("/status", methods=["GET"])
def status_page():
    logger.info(
        f"{request.remote_addr} {request.method} {request.url} {request.data}")
    return jsonify(Status())


def start():
    """
    Start the API server.
    """

    parser = argparse.ArgumentParser(
        description=
        "This is an open source decentralized application network. In this network, you can develop and publish decentralized applications."
    )

    parser.add_argument("-p",
                        "--port",
                        default=8000,
                        type=int,
                        help="Add new UNL node")

    args = parser.parse_args()

    logger.info(f"Starting API on port {args.port}")
    serve(app, host="0.0.0.0", port=args.port)


if __name__ == "__main__":
    start()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import os
import pickle
import time

from accounts.account import Account
from accounts.get_accounts import GetAccounts
from accounts.save_accounts import save_accounts
from app.app_main import app_tigger
from blockchain.block.blocks_hash import GetBlockshash
from blockchain.block.blocks_hash import SaveBlockshash
from blockchain.block.blocks_hash import SaveBlockshash_part
from blockchain.block.save_block_to_blockchain_db import \
    saveBlockstoBlockchainDB
from config import TEMP_BLOCK_PATH
from consensus.consensus_main import consensus_trigger
from lib.config_system import get_config
from lib.log import get_logger
from lib.perpetualtimer import perpetualTimer
from node.unl import Unl
from transactions.pending_to_validating import PendinttoValidating
from transactions.save_to_my_transaction import SavetoMyTransaction
from wallet.wallet import PrivateKey
from wallet.wallet import PublicKey
from wallet.wallet import Signature
from wallet.wallet import Wallet_Import

logger = get_logger("BLOCKCHAIN")


class Block:
    """
    Block class is most important class. It is responsible for
    resetting and saving blocks.

    You must give a creator of the block. This creator will
    own all the coins.
    """

    def __init__(
        self,
        creator,
        previous_hash="fb8b69c2276c8316c64a5d34b5f3063d1f8b8dc17cda7ee84fa1343978d464a9",
    ):
        self.genesis_time = int(time.time())
        self.start_time = int(time.time())
        self.block_time = 10

        self.newly = False

        self.previous_hash = previous_hash
        self.sequance_number = 0
        self.empty_block_number = 0

        blocks_hash = [self.previous_hash]
        SaveBlockshash(blocks_hash)

        accounts_list = GetAccounts()
        if accounts_list == []:
            save_accounts([Account(creator, 1000000000)])

        self.edited_accounts = []

        self.pendingTransaction = []
        self.validating_list = []
        self.transaction_fee = 0.02
        self.default_transaction_fee = 0.02
        # Each user settings by our hardware
        self.default_optimum_transaction_number = 10
        self.default_increase_of_fee = 0.01

        self.hash = None

        self.max_tx_number = 2
        self.minumum_transfer_amount = 1000

        self.raund_1_starting_time = None
        self.raund_1_time = 3.5
        self.raund_1 = False

        self.raund_2_starting_time = None
        self.raund_2_time = 3.5
        self.raund_2 = False

        self.consensus_timer = 0.50

        self.validated = False
        self.validated_time = None

        self.dowload_true_block = ""

        self.save_block()

        logger.info("Consensus timer is started")
        perpetualTimer(self.consensus_timer, consensus_trigger).start()

    def reset_the_block(self):
        """
        When the block is verified and if block have a transaction
        and if block have at least half of the max_tx_number transaction,it saves the block
        and makes the edits for the new block.
        """


        self.start_time = int(time.time())

        self.raund_1_starting_time = None
        self.raund_1 = False

        self.raund_2_starting_time = None
        self.raund_2 = False

        self.validated = False

        # Resetting the node candidate blocks.
        for node in Unl.get_as_node_type(Unl.get_unl_nodes()):
            node.candidate_block = None
            node.candidate_block_hash = None

        if not len(self.validating_list) == 0 and not len(
                self.validating_list) < (self.max_tx_number / 2):

            app_tigger(self)

            my_address = Wallet_Import(-1, 3)
            for tx in self.validating_list:
                if tx.toUser == my_address:
                    SavetoMyTransaction(tx)

            saveBlockstoBlockchainDB(self)

            # Resetting and setting the new elements.
            self.previous_hash = self.hash
            current_blockshash_list = GetBlockshash()
            current_blockshash_list.append(self.previous_hash)
            SaveBlockshash(current_blockshash_list)
            self.sequance_number = self.sequance_number + 1
            self.validating_list = []
            self.hash = None

            logger.info("New block created")
        else:
            logger.info(
                "New block not created because any transaction is not validated"
            )
            self.empty_block_number += 1

        logger.debug(self.__dict__)

        # Adding self.pendingTransaction to the new/current block.
        PendinttoValidating(self)

        # Saving the new block.
        self.save_block()

    def save_block(self):
        """
        Saves the current block to the TEMP_BLOCK_PATH.
        """

        os.chdir(get_config()["main_folder"])
        with open(TEMP_BLOCK_PATH, "wb") as block_file:
            pickle.dump(self, block_file, protocol=2)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import json
import os
import socket
import sys
import threading
import time

from config import *
from lib.log import get_logger
from node.unl import Unl

logger = get_logger("NODE")


class Node_Connection(threading.Thread):

    def __init__(self, main_node, sock, id, host, port):
        super(Node_Connection, self).__init__()

        self.host = host
        self.port = port
        self.main_node = main_node
        self.sock = sock
        self.terminate_flag = threading.Event()

        self.id = id

        self.candidate_block = None
        self.candidate_block_hash = None

        self.EOT_CHAR = 0x04.to_bytes(1, "big")

        from node.node import Node

        Node.save_connected_node(host, port, id)

    def send(self, data, encoding_type="utf-8"):

        if isinstance(data, str):
            self.sock.sendall(data.encode(encoding_type) + self.EOT_CHAR)

        elif isinstance(data, dict):
            try:
                json_data = json.dumps(data)
                json_data = json_data.encode(encoding_type) + self.EOT_CHAR
                self.sock.sendall(json_data)

            except TypeError as type_error:
                logger.exception(type_error)

            except Exception as e:
                logger.exception(e)

        elif isinstance(data, bytes):
            bin_data = data + self.EOT_CHAR
            self.sock.sendall(bin_data)

        else:
            logger.warning(
                "Node System: Node System: Datatype used is not valid please use str, dict (will be send as json) or bytes"
            )

    def stop(self):
        self.terminate_flag.set()

    def parse_packet(self, packet):
        try:
            packet_decoded = packet.decode("utf-8")

            try:
                return json.loads(packet_decoded)

            except json.decoder.JSONDecodeError:
                return packet_decoded

        except UnicodeDecodeError:
            return packet

    def run(self):
        self.sock.settimeout(10.0)
        buffer = b""

        while not self.terminate_flag.is_set():
            chunk = b""

            try:
                chunk = self.sock.recv(4096)

            except socket.timeout:
                logger.exception("Node System: Node_Connection: timeout")

            except Exception as e:
                self.terminate_flag.set()
                logger.exception(e)

            # BUG: possible buffer overflow when no EOT_CHAR is found => Fix by max buffer count or so?
            if chunk != b"":
                buffer += chunk
                eot_pos = buffer.find(self.EOT_CHAR)

                while eot_pos > 0:
                    packet = buffer[:eot_pos]
                    buffer = buffer[eot_pos + 1:]

                    self.main_node.message_from_node(self,
                                                     self.parse_packet(packet))

                    eot_pos = buffer.find(self.EOT_CHAR)

            time.sleep(0.01)

        # IDEA: Invoke (event) a method in main_node so the user is able to send a bye message to the node before it is closed?

        self.sock.settimeout(None)
        self.sock.close()
        logger.info("Node System: Node_Connection: Stopped")

    @staticmethod
    def connect(ip, port):
        """
        Connects to a node.
        """
        from node.node import Node

        Node.main_node.connect_to_node(ip, port)

    @staticmethod
    def connectmixdb():
        """
        Connects to nodes from mixdb.
        """
        from node.node import Node

        Node.connectionfrommixdb()

    def __str__(self):
        return f"Node Connection: {self.host}:{self.port} ({self.id})"



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from accounts.get_accounts import GetAccounts
from accounts.get_accounts_part import GetAccounts_part
from accounts.save_accounts_part import save_accounts_part
from blockchain.block.blocks_hash import GetBlockshash
from blockchain.block.blocks_hash import GetBlockshash_part
from blockchain.block.blocks_hash import SaveBlockshash_part
from lib.merkle_root import MerkleTree


def CalculateHash(block):
    """
    Calculates and returns the hash of the block.
    """

    # Transaction
    tx_list = []
    for element in block.validating_list[:]:
        tx_list.append(element.signature)
    if len(tx_list) != 0:
        tx_hash = MerkleTree(tx_list).getRootHash()
    else:
        tx_hash = "0"

    part_amount = 100000

    # Blocks Hash
    blocks_hash_list = []

    part_of_blocks_hash = GetBlockshash_part()
    the_blocks_hash = GetBlockshash()

    if (not len(the_blocks_hash) -
        (len(part_of_blocks_hash) * part_amount) == part_amount):
        for will_added_blocks_hash in the_blocks_hash[(
                len(part_of_blocks_hash) * part_amount):]:
            blocks_hash_list.append(will_added_blocks_hash)
    else:
        part_of_blocks_hash.append(
            MerkleTree(the_blocks_hash[(len(part_of_blocks_hash) *
                                        part_amount):]).getRootHash())
        SaveBlockshash_part(part_of_blocks_hash)

    for part_of_blocks_hash_element in part_of_blocks_hash:
        blocks_hash_list.append(part_of_blocks_hash_element)

    blockshash_hash = MerkleTree(blocks_hash_list).getRootHash()

    # Account
    account_list = []

    part_of_account = GetAccounts_part()
    the_accounts = GetAccounts()

    if not len(the_accounts) - (len(part_of_account) *
                                part_amount) >= part_amount:
        for will_added_accounts in the_accounts[(len(part_of_account) *
                                                 part_amount):]:
            account_list.append(str(will_added_accounts.dump_json()))
    else:
        part_of_account.append(
            MerkleTree(the_accounts[(len(part_of_account) *
                                     part_amount):]).getRootHash())
        save_accounts_part(part_of_account)

    for part_of_account_element in part_of_account:
        account_list.append(part_of_account_element)

    for edited_account in block.edited_accounts:
        account_list.append(str(edited_account.dump_json()))

    block.edited_accounts.clear()

    ac_hash = MerkleTree(account_list).getRootHash()

    # Other elements
    main_list = []
    main_list.append(block.previous_hash)
    main_list.append(str(block.sequance_number))
    main_list.append(blockshash_hash)
    main_list.append(ac_hash)
    main_list.append(tx_hash)
    main_list.append(str(block.default_transaction_fee))
    main_list.append(str(block.default_increase_of_fee))
    main_list.append(str(block.default_optimum_transaction_number))

    # Calculating and returning
    return MerkleTree(main_list).getRootHash()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


import pickle
import os

from lib.config_system import get_config

from config import TEMP_ACCOUNTS_PART_PATH


def save_accounts_part(the_accounts):
    """
    Saves the accounts to the TEMP_ACCOUNTS_PART_PATH.
    """

    os.chdir(get_config()["main_folder"])
    with open(TEMP_ACCOUNTS_PART_PATH, "wb") as block_file:
        pickle.dump(the_accounts, block_file, protocol=2)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from blockchain.block.block_main import Block
from blockchain.block.get_block import GetBlock
from lib.log import get_logger
from lib.settings_system import the_settings
from node.node import Node
from wallet.wallet import Wallet_Import

logger = get_logger("BLOCKCHAIN")


def CreateBlock():
    """
    If test mode is on, creates genesis block
    and send the connected nodes, if it is off,
    it calls get_block() function.
    """

    if the_settings()["test_mode"]:
        previous_hash = "0"

        try:
            current_block = GetBlock()
            if not current_block.hash is None:
                previous_hash = current_block.hash
            else:
                previous_hash = current_block.previous_hash
        except:
            pass

        logger.info(
            "Creating the genesis block and sending it to the connected nodes")
        Block(Wallet_Import(-1, 3), previous_hash)
        Node.main_node.send_full_accounts()
        Node.main_node.send_full_chain()
        Node.main_node.send_full_blockshash()
    else:
        logger.info("Getting the last block from the connected nodes")
        GetBlock()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from accounts.account import Account
from accounts.account import GetAccounts
from accounts.save_accounts import save_accounts
from wallet.wallet import Address


def ProccesstheTransaction(block):
    """
    It performs the transactions in the block.vali list and
    puts the transactions in order.

    Queuing is required so that all nodes have the same transaction hash.
    """

    from_user_list = []
    temp_validating_list = block.validating_list

    temp_accounts = GetAccounts()
    new_added_accounts_list = []

    for trans in block.validating_list:
        touser_inlist = False
        address_of_fromUser = Address(trans.fromUser)

        for Accounts in temp_accounts:

            if Accounts.Address == address_of_fromUser:
                Accounts.balance -= float(trans.amount) + trans.transaction_fee
                Accounts.sequance_number += 1
                from_user_list.append(Accounts)
                block.edited_accounts.append(Accounts)

            elif Accounts.Address == trans.toUser:
                Accounts.balance += float(trans.amount)
                touser_inlist = True
                block.edited_accounts.append(Accounts)

        # If not included in the temp_accounts, add.
        if not touser_inlist:
            new_added_accounts_list.append(
                Account(trans.toUser, float(trans.amount)))

    # Syncs new sorted list to block.validating_list

    block.validating_list = sorted(temp_validating_list,
                                   key=lambda x: x.fromUser)

    new_accounts_list = temp_accounts + sorted(new_added_accounts_list,
                                               key=lambda x: x.Address)

    save_accounts(new_accounts_list)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


def TXAlreadyGot(block, transaction):
    """
    Checks if the transaction is already in the block.
    """

    for already_tx in block.pendingTransaction + block.validating_list:
        if already_tx.signature == transaction.signature:
            return True

    return False



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


def ChangeTransactionFee(block):
    """
    Increase transaction fee by 0.01 DNC for each block.default_optimum_transaction_number argument
    """
    if not (len(block.pendingTransaction + block.validating_list) // block.default_optimum_transaction_number) == 0:
        increase = (len(block.pendingTransaction + block.validating_list) // block.default_optimum_transaction_number) * block.default_increase_of_fee
        block.transaction_fee += increase
    else:
        block.transaction_fee = block.default_transaction_fee



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import time

from blockchain.block.get_block import GetBlock
from consensus.consensus_first_round import consensus_round_1
from consensus.consensus_second_round import consensus_round_2
from lib.log import get_logger

logger = get_logger("CONSENSUS")


def consensus_trigger():
    """
    Consensus process consists of 2 stages. This function makes
    the necessary redirects according to the situation and works
    to shorten the block time.
    """

    block = GetBlock()

    logger.info(
        f"BLOCK#{block.sequance_number}:{block.empty_block_number} Consensus process started"
    )

    if block.validated:
        true_time = (block.genesis_time + block.block_time +
                     ((block.sequance_number) * block.block_time))
        if block.newly:
            true_time -= 1
            logger.info(
                "Consensus proccess is complated but the time is not true, will be waiting for the true time"
            )
        if not int(time.time()) < true_time:
            block.newly = False
            logger.info(
                "Consensus proccess is complated, the block will be reset")
            block.reset_the_block()
    else:
        if block.raund_1_starting_time is None:
            block.raund_1_starting_time = int(time.time())
            block.save_block()
        if not block.raund_1:
            logger.info("First round is starting")
            consensus_round_1(block)
            logger.info("First round is done")
        elif not block.raund_2:
            logger.info("Second round is starting")
            consensus_round_2(block)
            logger.info("Second round is done")

    logger.info("Consensus process is done")



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import time

from accounts.get_balance import GetBalance
from accounts.get_sequance_number import GetSequanceNumber
from lib.log import get_logger
from transactions.change_transaction_fee import ChangeTransactionFee
from transactions.propagating_the_tx import PropagatingtheTX
from transactions.transaction import Transaction
from transactions.tx_already_got import TXAlreadyGot
from wallet.wallet import Ecdsa
from wallet.wallet import PublicKey
from wallet.wallet import Signature

logger = get_logger("TRANSACTIONS")


def CheckTransaction(block, transaction):
    """
    This function checks the transaction.
    """

    logger.info(
        f"Checking the transaction started {block.sequance_number}:{transaction.signature}"
    )

    validation = True

    if not TXAlreadyGot(block, transaction):
        logger.warning("The transaction is already got")
    else:
        validation = False

    if Ecdsa.verify(
        (str(transaction.sequance_number) + str(transaction.fromUser) +
         str(transaction.toUser) + str(transaction.data) +
         str(transaction.amount) + str(transaction.transaction_fee) +
         str(transaction.transaction_time)),
            Signature.fromBase64(transaction.signature),
            PublicKey.fromPem(transaction.fromUser),
    ):
        logger.info("The signature is valid")
    else:
        validation = False

    if not transaction.amount < block.minumum_transfer_amount:
        logger.info("Minimum transfer amount is reached")
    else:
        validation = False

    if not transaction.transaction_fee < block.transaction_fee:
        logger.info("Transaction fee is reached")
    else:
        validation = False

    if not (int(time.time()) - transaction.transaction_time) > 60:
        logger.info("Transaction time is valid")
    else:
        validation = False

    if transaction.sequance_number == (
            GetSequanceNumber(transaction.fromUser) + 1):
        logger.info("Sequance number is valid")
    else:
        validation = False

    balance = GetBalance(block, transaction.fromUser)
    if balance >= (float(transaction.amount) +
                   float(transaction.transaction_fee)):
        logger.info("Balance is valid")
    else:
        validation = False

    if (balance -
        (float(transaction.amount) + float(transaction.transaction_fee))) > 2:
        logger.info("Balance is enough")
    else:
        validation = False

    for tx in block.pendingTransaction + block.validating_list:
        if tx.fromUser == transaction.fromUser and tx.signature != transaction.signature:
            logger.info("Multiple transaction in one account")
            validation = False

    logger.info(
        f"Checking the transaction finished {block.sequance_number}:{transaction.signature}"
    )
    return validation



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


def PendinttoValidating(block):
    """
    Adds transactions to the verification list
    if there are suitable conditions.
    """

    if (len(block.validating_list) < block.max_tx_number
            and block.raund_1_starting_time is None):
        for tx in block.pendingTransaction[:]:
            if len(block.validating_list) < block.max_tx_number:
                block.validating_list.append(tx)
                block.pendingTransaction.remove(tx)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import time

from blockchain.block.get_block import GetBlock
from node.unl import Unl
from transactions.get_my_transaction import GetMyTransaction


def Status():
    """
    Returns the status of the network.
    """

    first_block = GetBlock()
    start_time = time.time()
    while True:

        time.sleep(15)
        new_time = time.time()
        new_block = GetBlock()

        status_json = {
            "status":
            "",
            "first_block":
            str(first_block.__dict__),
            "new_block":
            str(new_block.__dict__),
            "last_transaction_of_block":
            str(new_block.validating_list[-1])
            if len(new_block.validating_list) > 0 else "",
            "last_transaction_of_us":
            str([i.__dict__ for i in GetMyTransaction()]),
            "connected_nodes": [
                str(the_connections) for the_connections in
                Unl.get_as_node_type(Unl.get_unl_nodes())
            ],
        }

        if not (first_block.sequance_number +
                first_block.empty_block_number) == (
                    new_block.sequance_number + new_block.empty_block_number):
            status_json["status"] = "Working"
        else:
            status_json["status"] = "Not working"

        return status_json



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from accounts.get_accounts import GetAccounts
from wallet.wallet import Address


def GetBalance(block, user):
    """
    Returns the users balance.
    """

    balance = -block.minumum_transfer_amount
    user = Address(user)
    for Accounts in GetAccounts():

        if Accounts.Address == user:
            balance += Accounts.balance
            break

    return balance



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import time

from blockchain.block.calculate_hash import CalculateHash
from blockchain.candidate_block.get_candidate_blocks import GetCandidateBlocks
from lib.log import get_logger
from node.node import Node
from node.unl import Unl
from transactions.process_the_transaction import ProccesstheTransaction
from transactions.send_transaction_to_the_block import \
    SendTransactiontoTheBlock

logger = get_logger("CONSENSUS_FIRST_ROUND")


def consensus_round_1(block):
    """
    At this stage of the consensus process,
    The transactions of our and the unl nodes
    are evaluated and transactions accepted by
    owned by more than 50 percent.
    Inputs:
      * block: The block (class) we want consensus
      round 1 to be done
    """

    logger.info(
        f"BLOCK#{block.sequance_number}:{block.empty_block_number} First round is starting"
    )

    unl_nodes = Unl.get_unl_nodes()
    logger.info("Our block is sending to the unl nodes")
    Node.main_node.send_my_block(block, Unl.get_as_node_type(unl_nodes))
    candidate_class = GetCandidateBlocks()
    time_difference = int(time.time()) - block.raund_1_starting_time
    if len(candidate_class.candidate_blocks) > ((len(unl_nodes) * 80) / 100):
        logger.info("Enough candidate blocks received")

        logger.info("Time difference is {}".format(time_difference))
        logger.info("block.raund_1_time is {}".format(block.raund_1_time))

        if time_difference > block.raund_1_time:
            logger.info("True time")
            temp_validating_list = []
            for candidate_block in candidate_class.candidate_blocks[:]:
                logger.debug(f"Candidate block {str(candidate_block)}")

                for other_block_tx in candidate_block["transaction"]:

                    tx_valid = 0

                    for my_txs in block.validating_list:
                        if other_block_tx.signature == my_txs.signature:
                            tx_valid += 1

                    if len(candidate_class.candidate_blocks) != 1:

                        for other_block in candidate_class.candidate_blocks[:]:
                            if candidate_block["signature"] != other_block[
                                    "signature"]:

                                for other_block_txs in other_block[
                                        "transaction"]:
                                    if (other_block_tx.signature ==
                                            other_block_txs.signature):

                                        tx_valid += 1
                    else:
                        tx_valid += 1

                    logger.debug(
                        f"Tx valid of {other_block_tx.signature} : {tx_valid}")
                    if tx_valid > (len(unl_nodes) / 2):

                        already_in_ok = False
                        for alrady_tx in temp_validating_list[:]:

                            if other_block_tx.signature == alrady_tx.signature:
                                logger.warning(
                                    "The transaction is already in the list")
                                already_in_ok = True
                        if not already_in_ok:
                            logger.info(
                                f"Transaction is valid ({other_block_tx.signature})"
                            )
                            temp_validating_list.append(other_block_tx)

            newly_added_list = []

            for my_validating_list in block.validating_list[:]:
                ok = False
                for my_temp_validating_list in temp_validating_list[:]:
                    if (my_validating_list.signature ==
                            my_temp_validating_list.signature):
                        ok = True
                block.validating_list.remove(my_validating_list)
                if not ok:
                    newly_added_list.append(my_validating_list)

            block.validating_list = temp_validating_list
            logger.debug(f"Newly validating list {block.validating_list}")

            for each_newly in newly_added_list:
                SendTransactiontoTheBlock(
                    block,
                    each_newly.sequance_number,
                    each_newly.signature,
                    each_newly.fromUser,
                    each_newly.toUser,
                    each_newly.transaction_fee,
                    each_newly.data,
                    each_newly.amount,
                    transaction_sender=None,
                    transaction_time=each_newly.transaction_time,
                )

            block.raund_1 = True

            block.raund_2_starting_time = int(time.time())

            ProccesstheTransaction(block)

            block.hash = CalculateHash(block)
            logger.debug(f"Block hash {block.hash}")

            block.save_block()

    logger.info("First round is done")



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import json
import os
import socket
import sys
import threading
import time
from hashlib import sha256

from blockchain.block.get_block import GetBlock
from config import CONNECTED_NODE_PATH
from config import LOADING_BLOCK_PATH
from config import TEMP_ACCOUNTS_PATH
from config import TEMP_BLOCK_PATH
from config import TEMP_BLOCKSHASH_PATH
from lib.log import get_logger
from lib.merkle_root import MerkleTree
from node.node import *
from node.node_connection import Node_Connection
from node.unl import Unl
from transactions.transaction import Transaction
from wallet.wallet import Ecdsa
from wallet.wallet import PrivateKey
from wallet.wallet import PublicKey
from wallet.wallet import Signature
from wallet.wallet import Wallet_Import

logger = get_logger("NODE")


class Node(threading.Thread):

    main_node = None
    unl_nodes = []

    id = "".join([
        l.strip() for l in Wallet_Import(0, 0).splitlines()
        if l and not l.startswith("-----")
    ])

    def __init__(self, host, port, callback=None):
        self.__class__.main_node = self
        super(Node, self).__init__()

        self.terminate_flag = threading.Event()

        self.host = host
        self.port = port

        self.callback = callback

        self.nodes_inbound = []

        self.nodes_outbound = []

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.init_server()

        self.start()

    def run(self):

        while not self.terminate_flag.is_set():
            try:
                connection, client_address = self.sock.accept()

                connected_node_id = connection.recv(4096).decode("utf-8")
                connection.send(Node.id.encode("utf-8"))
                if Unl.node_is_unl(connected_node_id):
                    thread_client = self.create_the_new_connection(
                        connection,
                        connected_node_id,
                        client_address[0],
                        client_address[1],
                    )
                    thread_client.start()

                    self.nodes_inbound.append(thread_client)
                else:
                    logger.warning(
                        "Node System: Could not connect with node because node is not unl node."
                    )

            except socket.timeout:
                pass

            except Exception as e:
                raise e

            time.sleep(0.01)

        logger.info("Node System: Stopping protocol started by node")
        for t in self.nodes_inbound:
            t.stop()

        for t in self.nodes_outbound:
            t.stop()

        time.sleep(1)

        for t in self.nodes_inbound:
            t.join()

        for t in self.nodes_outbound:
            t.join()

        self.sock.settimeout(None)
        self.sock.close()
        logger.info("Node System: The node is stopped")

    def init_server(self):
        logger.info("Node System: Node server is starting")
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind((self.host, self.port))
        self.sock.settimeout(10.0)
        self.sock.listen(1)

    def delete_closed_connections(self):

        for n in self.nodes_inbound:
            if n.terminate_flag.is_set():
                n.join()
                del self.nodes_inbound[self.nodes_inbound.index(n)]

        for n in self.nodes_outbound:
            if n.terminate_flag.is_set():
                n.join()
                del self.nodes_outbound[self.nodes_inbound.index(n)]

    def send_data_to_nodes(self, data, exclude=[]):

        for n in self.nodes_inbound:
            if n in exclude:
                logger.info(
                    "Node System: Node send_data_to_nodes: Node is excluded")
            else:
                try:
                    self.send_data_to_node(n, data)
                except:  # lgtm [py/catch-base-exception]
                    pass

        for n in self.nodes_outbound:
            if n in exclude:
                logger.info(
                    "Node System: Node send_data_to_nodes: Node is excluded")
            else:
                try:
                    self.send_data_to_node(n, data)
                except:  # lgtm [py/catch-base-exception]
                    pass

    def send_data_to_node(self, n, data):

        self.delete_closed_connections()
        if n in self.nodes_inbound or n in self.nodes_outbound:
            try:
                n.send(data)

            except Exception as e:
                logger.exception(
                    "Node System: Node send_data_to_node: Could not send data to node"
                )
        else:
            logger.warning(
                "Node System: Node send_data_to_node: Node is not connected")

    def connect_to_node(self, host, port):

        if host == self.host and port == self.port:
            logger.error(
                "Node System: Node connect_to_node: You can not connect to yourself"
            )
            return False

        for node in self.nodes_outbound:
            if node.host == host and node.port == port:
                logger.warning(
                    "Node System: connect_to_node: Node is already connected")
                return True

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            logger.info("Node System: Connecting to %s port %s" % (host, port))
            sock.connect((host, port))

            # Basic information exchange (not secure) of the id's of the nodes!
            # Send my id to the connected node!
            sock.send(Node.id.encode("utf-8"))
            # When a node is connected, it sends it id!
            connected_node_id = sock.recv(4096).decode("utf-8")

            if Unl.node_is_unl(connected_node_id):
                thread_client = self.create_the_new_connection(
                    sock, connected_node_id, host, port)
                thread_client.start()

                self.nodes_outbound.append(thread_client)
            else:
                logger.warning(
                    "Node System: Could not connect with node because node is not unl node."
                )
        except Exception as e:
            logger.exception("Node System: Could not connect with node")

    def disconnect_to_node(self, node):

        if node in self.nodes_outbound:
            node.stop()
            node.join()
            del self.nodes_outbound[self.nodes_outbound.index(node)]

        else:
            print(
                "Node System: Node disconnect_to_node: Node is not connected")

    def stop(self):
        self.terminate_flag.set()

    def create_the_new_connection(self, connection, id, host, port):

        return Node_Connection(self, connection, id, host, port)

    @staticmethod
    def get_connected_node():
        """
        Returns the connected nodes.
        """

        sys.path.append(os.path.join(os.path.dirname(__file__), ".."))

        from lib.config_system import get_config

        if not os.path.exists(CONNECTED_NODE_PATH):
            temp_json = {}
            return temp_json

        os.chdir(get_config()["main_folder"])
        with open(CONNECTED_NODE_PATH, "rb") as connected_node_file:
            return json.load(connected_node_file)

    @staticmethod
    def save_connected_node(host, port, id):
        """
        Saves the connected nodes.
        """

        node_list = Node.get_connected_node()

        already_in_list = False

        for element in node_list:
            if (node_list[element]["host"] == host
                    and node_list[element]["port"] == port):
                already_in_list = True

        if not already_in_list:
            node_list[id] = {}
            node_list[id]["host"] = host
            node_list[id]["port"] = port

            sys.path.append(os.path.join(os.path.dirname(__file__), ".."))

            from lib.config_system import get_config

            os.chdir(get_config()["main_folder"])
            with open(CONNECTED_NODE_PATH, "w") as connected_node_file:
                json.dump(node_list, connected_node_file, indent=4)

    @staticmethod
    def connectionfrommixdb():
        """
        Connects to the mixdb.
        """

        node_list = Node.get_connected_node()
        from node.node import Node

        for element in node_list:
            Node.main_node.connect_to_node(node_list[element]["host"],
                                           node_list[element]["port"])

    @staticmethod
    def connected_node_delete(node):
        """
        Deletes a connected node.
        """

        saved_nodes = Node.get_connected_node()
        if node in saved_nodes:
            del saved_nodes[node]
            from lib.config_system import get_config

            os.chdir(get_config()["main_folder"])
            with open(CONNECTED_NODE_PATH, "w") as connected_node_file:
                json.dump(saved_nodes, connected_node_file, indent=4)

    def message_from_node(self, node, data):

        if str(data) == "sendmefullblock":
            self.send_full_chain(node)

        try:
            if (data["fullblock"] == 1 and Unl.node_is_unl(node.id)
                    and Ecdsa.verify(
                        "fullblock" + data["byte"],
                        Signature.fromBase64(data["signature"]),
                        PublicKey.fromPem(node.id),
                    )):
                print("getting chain")
                self.get_full_chain(data, node)
        except Exception as e:
            print(e)

        try:

            if (data["fullaccounts"] == 1 and Unl.node_is_unl(node.id)
                    and Ecdsa.verify(
                        "fullaccounts" + data["byte"],
                        Signature.fromBase64(data["signature"]),
                        PublicKey.fromPem(node.id),
                    )):
                print("getting chain")
                self.get_full_accounts(data, node)
        except Exception as e:
            print(e)

        try:

            if (data["fullblockshash"] == 1 and Unl.node_is_unl(node.id)
                    and Ecdsa.verify(
                        "fullblockshash" + data["byte"],
                        Signature.fromBase64(data["signature"]),
                        PublicKey.fromPem(node.id),
                    )):
                self.get_full_blockshash(data, node)
        except Exception as e:
            print(e)

        try:
            if data["transactionrequest"] == 1:
                self.get_transaction(data, node)
        except Exception as e:
            print(e)

        try:
            if data["action"] == "myblock":
                self.get_candidate_block(data, node)
        except Exception as e:
            print(e)

        try:
            if data["action"] == "myblockhash":
                self.get_candidate_block_hash(data, node)
        except Exception as e:
            print(e)

    def send_my_block(self, block, nodes):
        system = block

        new_list = []

        signature_list = []

        for element in system.validating_list:
            new_list.append(element.dump_json())
            signature_list.append(element.signature)

        Merkle_signature_list = (MerkleTree(signature_list).getRootHash()
                                 if len(signature_list) != 0 else "0")

        data = {
            "action":
            "myblock",
            "transaction":
            new_list,
            "sequance_number":
            system.sequance_number,
            "signature":
            Ecdsa.sign(
                "myblock" + Merkle_signature_list +
                str(system.sequance_number),
                PrivateKey.fromPem(Wallet_Import(0, 1)),
            ).toBase64(),
        }

        for each_node in nodes:
            self.send_data_to_node(each_node, data)

    def send_my_block_hash(self, block, nodes):
        system = block

        if system.raund_1 and not system.raund_2:

            data = {
                "action":
                "myblockhash",
                "hash":
                system.hash,
                "sequance_number":
                system.sequance_number,
                "signature":
                Ecdsa.sign(
                    "myblockhash" + system.hash + str(system.sequance_number),
                    PrivateKey.fromPem(Wallet_Import(0, 1)),
                ).toBase64(),
            }

            for each_node in nodes:
                self.send_data_to_node(each_node, data)

    def get_candidate_block(self, data, node):
        if node.candidate_block is None:

            if (Unl.node_is_unl(node.id)
                    and GetBlock().sequance_number == data["sequance_number"]):

                signature_list = []
                for element in data["transaction"]:
                    signature_list.append(element["signature"])

                merkle_root_of_signature_list = (
                    MerkleTree(signature_list).getRootHash()
                    if len(signature_list) != 0 else "0")

                if Ecdsa.verify(
                        "myblock" + merkle_root_of_signature_list +
                        str(data["sequance_number"]),
                        Signature.fromBase64(data["signature"]),
                        PublicKey.fromPem(node.id),
                ):

                    temp_tx = []

                    for element in data["transaction"]:
                        temp_tx.append(Transaction.load_json(element))

                    data["transaction"] = temp_tx

                    node.candidate_block = data

    def get_candidate_block_hash(self, data, node):

        if node.candidate_block_hash is None:

            if (Unl.node_is_unl(node.id)
                    and GetBlock().sequance_number == data["sequance_number"]):

                if Ecdsa.verify(
                        "myblockhash" + data["hash"] +
                        str(data["sequance_number"]),
                        Signature.fromBase64(data["signature"]),
                        PublicKey.fromPem(node.id),
                ):
                    data["sender"] = node.id

                    node.candidate_block_hash = data

    def send_full_chain(self, node=None):
        file = open(TEMP_BLOCK_PATH, "rb")
        SendData = file.read(1024)
        while SendData:

            data = {
                "fullblock":
                1,
                "byte": (SendData.decode(encoding="iso-8859-1")),
                "signature":
                Ecdsa.sign(
                    "fullblock" + str(
                        (SendData.decode(encoding="iso-8859-1"))),
                    PrivateKey.fromPem(Wallet_Import(0, 1)),
                ).toBase64(),
            }
            if not node is None:
                self.send_data_to_node(node, data)
            else:
                self.send_data_to_nodes(data)

            SendData = file.read(1024)

            if not SendData:
                data = {
                    "fullblock":
                    1,
                    "byte":
                    "end",
                    "signature":
                    Ecdsa.sign("fullblock" + "end",
                               PrivateKey.fromPem(Wallet_Import(
                                   0, 1))).toBase64(),
                }
                if not node is None:
                    self.send_data_to_node(node, data)
                else:
                    self.send_data_to_nodes(data)

    def send_full_accounts(self, node=None):
        file = open(TEMP_ACCOUNTS_PATH, "rb")
        SendData = file.read(1024)
        while SendData:

            data = {
                "fullaccounts":
                1,
                "byte": (SendData.decode(encoding="iso-8859-1")),
                "signature":
                Ecdsa.sign(
                    "fullaccounts" + str(
                        (SendData.decode(encoding="iso-8859-1"))),
                    PrivateKey.fromPem(Wallet_Import(0, 1)),
                ).toBase64(),
            }
            if not node is None:
                self.send_data_to_node(node, data)
            else:
                self.send_data_to_nodes(data)

            SendData = file.read(1024)

            if not SendData:
                data = {
                    "fullaccounts":
                    1,
                    "byte":
                    "end",
                    "signature":
                    Ecdsa.sign("fullaccounts" + "end",
                               PrivateKey.fromPem(Wallet_Import(
                                   0, 1))).toBase64(),
                }
                if not node is None:
                    self.send_data_to_node(node, data)
                else:
                    self.send_data_to_nodes(data)

    def send_full_blockshash(self, node=None):
        file = open(TEMP_BLOCKSHASH_PATH, "rb")
        SendData = file.read(1024)
        while SendData:

            data = {
                "fullblockshash":
                1,
                "byte": (SendData.decode(encoding="iso-8859-1")),
                "signature":
                Ecdsa.sign(
                    "fullblockshash" + str(
                        (SendData.decode(encoding="iso-8859-1"))),
                    PrivateKey.fromPem(Wallet_Import(0, 1)),
                ).toBase64(),
            }
            if not node is None:
                self.send_data_to_node(node, data)
            else:
                self.send_data_to_nodes(data)

            SendData = file.read(1024)

            if not SendData:
                data = {
                    "fullblockshash":
                    1,
                    "byte":
                    "end",
                    "signature":
                    Ecdsa.sign(
                        "fullblockshash" + "end",
                        PrivateKey.fromPem(Wallet_Import(0, 1)),
                    ).toBase64(),
                }
                if not node is None:
                    self.send_data_to_node(node, data)
                else:
                    self.send_data_to_nodes(data)

    def get_full_chain(self, data, node):

        get_ok = False

        if not os.path.exists(TEMP_BLOCK_PATH):
            get_ok = True
        else:
            system = GetBlock()
            if node.id == system.dowload_true_block:
                get_ok = True

        if get_ok:

            if str(data["byte"]) == "end":

                os.rename(LOADING_BLOCK_PATH, TEMP_BLOCK_PATH)

                from consensus.consensus_main import consensus_trigger
                from lib.perpetualtimer import perpetualTimer

                system = GetBlock()
                system.newly = True
                from transactions.change_transaction_fee import \
                    ChangeTransactionFee

                ChangeTransactionFee(system)

                system.exclude_validators = []
                perpetualTimer(system.consensus_timer,
                               consensus_trigger).start()
                system.save_block()

            else:
                file = open(LOADING_BLOCK_PATH, "ab")

                file.write((data["byte"].encode(encoding="iso-8859-1")))
                file.close()

    def get_full_blockshash(self, data, node):

        get_ok = False

        if not os.path.exists(TEMP_BLOCKSHASH_PATH):
            get_ok = True
        else:
            system = GetBlock()
            if node.id == system.dowload_true_block:
                get_ok = True

        if get_ok:
            file = open(TEMP_BLOCKSHASH_PATH, "ab")

            file.write((data["byte"].encode(encoding="iso-8859-1")))
            file.close()

    def get_full_accounts(self, data, node):

        get_ok = False

        if not os.path.exists(TEMP_ACCOUNTS_PATH):
            get_ok = True
        else:
            system = GetBlock()
            if node.id == system.dowload_true_block:
                get_ok = True

        if get_ok:
            file = open(TEMP_ACCOUNTS_PATH, "ab")

            file.write((data["byte"].encode(encoding="iso-8859-1")))
            file.close()

    def get_transaction(self, data, node):
        system = GetBlock()
        from transactions.send_transaction_to_the_block import \
            SendTransactiontoTheBlock

        SendTransactiontoTheBlock(
            system,
            sequance_number=data["sequance_number"],
            signature=data["signature"],
            fromUser=data["fromUser"],
            toUser=data["to_user"],
            data=data["data"],
            amount=data["amount"],
            transaction_fee=data["transaction_fee"],
            transaction_sender=node,
            transaction_time=data["transaction_time"],
        )



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import time
from hashlib import sha256

from accounts.get_sequance_number import GetSequanceNumber
from blockchain.block.get_block import GetBlock
from lib.settings_system import the_settings
from transactions.save_to_my_transaction import SavetoMyTransaction
from transactions.send_transaction_to_the_block import \
    SendTransactiontoTheBlock
from wallet.wallet import Ecdsa
from wallet.wallet import PrivateKey
from wallet.wallet import Wallet_Import


def send(password, to_user, amount, data=None):
    """
    The main function for sending the transaction.

    Inputs:
        password: The password of the wallet.
        to_user: The address of the recipient.
        amount: The amount of the transaction.
        data: The data of the transaction.

    """

    try:
        amount = float(amount)
    except ValueError:
        print("This is not float coin amount.")
        return False

    if isinstance(amount, int):
        amount = float(amount)

    if not isinstance(amount, float):
        print("This is not int or float coin amount.")
        return None

    if amount < 0:
        print("This is negative coin amount.")
        return None

    block = GetBlock()

    if not amount < block.minumum_transfer_amount:
        if (Wallet_Import(int(the_settings()["wallet"]),
                          2) == sha256(password.encode("utf-8")).hexdigest()):

            my_private_key = Wallet_Import(-1, 1, password)
            my_public_key = "".join([
                l.strip() for l in Wallet_Import(-1, 0).splitlines()
                if l and not l.startswith("-----")
            ])

            sequance_number = GetSequanceNumber(my_public_key) + 1

            # Get the current fee
            transaction_fee = block.transaction_fee

            tx_time = int(time.time())

            the_tx = SendTransactiontoTheBlock(
                block,
                sequance_number=sequance_number,
                signature=Ecdsa.sign(
                    str(sequance_number) + str(my_public_key) + str(to_user) +
                    str(data) + str(amount) + str(transaction_fee) +
                    str(tx_time),
                    PrivateKey.fromPem(my_private_key),
                ).toBase64(),
                fromUser=str(my_public_key),
                toUser=str(to_user),
                data=data,
                amount=amount,
                transaction_fee=transaction_fee,
                transaction_sender=None,
                transaction_time=tx_time,
            )

            if not the_tx == False:
                SavetoMyTransaction(the_tx)

            del my_private_key
            del password

        else:
            print("Password is not correct")



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import time

from accounts.get_balance import GetBalance
from accounts.get_sequance_number import GetSequanceNumber
from transactions.change_transaction_fee import ChangeTransactionFee
from transactions.check_transaction import CheckTransaction
from transactions.propagating_the_tx import PropagatingtheTX
from transactions.transaction import Transaction
from transactions.tx_already_got import TXAlreadyGot
from wallet.wallet import Ecdsa
from wallet.wallet import PublicKey
from wallet.wallet import Signature


def SendTransactiontoTheBlock(
    block,
    sequance_number,
    signature,
    fromUser,
    toUser,
    transaction_fee,
    data,
    amount,
    transaction_time,
    transaction_sender=None,
):
    """
    This function creates a transaction and adds it
    to the validating list and other direction.
    """

    the_tx = Transaction(
        sequance_number=sequance_number,
        signature=signature,
        fromUser=fromUser,
        toUser=toUser,
        data=data,
        amount=amount,
        transaction_fee=transaction_fee,
        time_of_transaction=transaction_time,
    )
    print(the_tx.dump_json())
    checking = CheckTransaction(block, the_tx)

    if checking:
        block.pendingTransaction.append(the_tx)
        ChangeTransactionFee(block)
        block.save_block()

        PropagatingtheTX(the_tx)

        return the_tx
    else:
        return False



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import time

from blockchain.candidate_block.get_candidate_blocks import GetCandidateBlocks
from lib.log import get_logger
from node.node import Node
from node.unl import Unl

logger = get_logger("CONSENSUS_SECOND_ROUND")


def consensus_round_2(block):
    """
    At this stage of the consensus process,
    The blocks (block hashes) of our and the
    unl nodes are compared and the block accepted
    by 80 percent is approved.

    Inputs:
      * block: The block (class) we want consensus
      round 2 to be done
    """

    logger.info(
        f"BLOCK#{block.sequance_number}:{block.empty_block_number} Second round is starting"
    )

    unl_nodes = Unl.get_unl_nodes()
    logger.info("Our block hash is sending to the unl nodes")
    Node.main_node.send_my_block_hash(block, Unl.get_as_node_type(unl_nodes))

    candidate_class = GetCandidateBlocks()

    time_difference = int(time.time()) - block.raund_2_starting_time

    if len(candidate_class.candidate_block_hashes) > (
        (len(unl_nodes) * 80) / 100):
        logger.info("Enough candidate block hashes received")

        if time_difference > block.raund_2_time:
            logger.info("True time")

            for candidate_block in candidate_class.candidate_block_hashes[:]:
                logger.debug(f"Candidate block hash {candidate_block}")

                tx_valid = 0

                if block.hash == candidate_block["hash"]:
                    tx_valid += 1

                for other_block in candidate_class.candidate_block_hashes[:]:

                    if candidate_block != other_block:
                        if candidate_block["hash"] == other_block["hash"]:
                            tx_valid += 1

                logger.debug(f"Hash valid of  {candidate_block} : {tx_valid}")
                if tx_valid > ((len(unl_nodes) * 80) / 100):

                    if block.hash == candidate_block["hash"]:
                        logger.info("Block approved")
                        block.validated = True
                        block.validated_time = int(time.time())
                        block.raund_2 = True

                    else:
                        sender = candidate_block["sender"]
                        logger.warning(
                            f"Our block is not valid, the system will try to get true block from node {sender}"
                        )
                        node = Node.main_node
                        unl_list = Unl.get_as_node_type([sender])
                        node.send_data_to_node(unl_list[0], "sendmefullblock")
                        block.dowload_true_block = sender
                    block.save_block()

    logger.info("Second round is done")



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from setuptools import setup

setup(
    name="decentra_network",
    version="0.21.0",
    description=
    """This is an open source decentralized application network. In this network, you can develop and publish decentralized applications.""",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    url="https://decentra-network.github.io/Decentra-Network/",
    author="Decentra Network Developers",
    author_email="atadogan06@gmail.com",
    license="MPL-2.0",
    packages=[
        "",
        "accounts",
        "app",
        "blockchain",
        "blockchain.block",
        "blockchain.candidate_block",
        "consensus",
        "gui_lib",
        "gui_lib.libs",
        "gui_lib.libs.baseclass",
        "lib",
        "node",
        "transactions",
        "wallet",
    ],
    include_package_data=True,
    package_dir={"": "src"},
    entry_points={
        "console_scripts":
        ["dngui = gui:start", "dncli=cli:start", "dnapi=api:start"],
    },
    python_requires=">=3.7",
    zip_safe=False,
)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import os
from threading import Thread

from lib.log import get_logger

logger = get_logger("APP")



class app(Thread):
    """
    It initiates the start functions of the applications in parallel
    from the main process.
    """

    def __init__(self, import_command, func, block):
        Thread.__init__(self)
        self.import_command = import_command
        self.func = func
        self.block = block

    def run(self):
        """
        Run the application.
        """
        exec(self.import_command)
        for trans in self.block.validating_list:  # lgtm [py/unused-loop-variable]
            logger.debug(f"Application triggering for tx {trans.__dict__}")
            exec(self.func)




def app_tigger(block):
    """
    Notifies applications of validated transactions after
    the block is validated.
    """

    logger.info(f"Triggering applications for block {block.sequance_number}:{block.empty_block_number}")

    for folder_entry in os.scandir("app"):
        logger.debug(f"Found application {folder_entry.name}")
        if (".md" not in folder_entry.name and "__" not in folder_entry.name
                and "app_main" not in folder_entry.name):
            logger.debug(f"Starting thread for application {folder_entry.name}")
            for entry in os.scandir("app/" + folder_entry.name):
                logger.debug(f"Found entry {entry.name}")
                if entry.is_file():
                    logger.debug(f"Found file {entry.name}")
                    if (entry.name[0] != "_" and ".py" in entry.name
                            and "_main" in entry.name):
                        logger.debug(f"Starting thread for file {entry.name}")
                        import_command = f"from app.{folder_entry.name}.{entry.name.replace('.py','')} import {entry.name.replace('.py','')}_tx"
                        tx_command = f"{entry.name.replace('.py','')}_tx(trans)"
                        logger.debug(f"import command: {import_command}")
                        logger.debug(f"tx_command: {tx_command}")
                        try:
                            x = app(import_command, tx_command, block)
                            x.start()
                        except Exception as e:
                            logger.exception(e)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from node.node import Node
from node.unl import Unl


def PropagatingtheTX(tx):
    """
    Sends the given transaction to UNL nodes.
    """

    items = {
        "transactionrequest": 1,
        "sequance_number": tx.sequance_number,
        "signature": tx.signature,
        "fromUser": tx.fromUser,
        "to_user": tx.toUser,
        "data": tx.data,
        "amount": tx.amount,
        "transaction_fee": tx.transaction_fee,
        "transaction_time": tx.transaction_time,
    }
    for each_node in Unl.get_as_node_type(Unl.get_unl_nodes()):
        Node.main_node.send_data_to_node(each_node, items)
